/*!
  \file model.h
  \brief Contains the Model Class that handles 3D-Models
*/

#ifndef _MODEL_H
#define _MODEL_H

#include "abstract_model.h"

#include "material.h"
#include "glincl.h"
#include "array.h"
#include <list>

// FORWARD DECLARATION //
template<class T> class tArray;

/////////////
/// MODEL ///
/////////////
//! Class that handles 3D-Models. it can also read them in and display them.
class Model : public AbstractModel
{
 public:
  Model(const char* modelName = NULL, MODEL_TYPE type = MODEL_DISPLAY_LIST);
  virtual ~Model();

  void draw() const;

  /** @returns a Pointer to the Vertex-Array, if it was deleted it returns NULL */
  inline const GLfloat* getVertexArray() const { return this->vertices->getArray(); };
  /** @returns the VertexCount of this Model */
  inline unsigned int getVertexCount() const { return this->vertexCount; };

  /** @returns a Pointer to the Normals-Array, if it was deleted it returns NULL */
  inline const GLfloat* getNormalsArray() const { return this->normals->getArray(); };
  /** @returns the NormalsCount of this Model */
  inline unsigned int getNormalsCount() const { return this->normalCount; };

  /** @returns a Pointer to the TexCoord-Array, if it was deleted it returns NULL */
  inline const GLfloat* getTexCoordArray() const { return this->vTexture->getArray(); };
  /** @returns the TexCoord-Count of this Model */
  inline unsigned int getTexCoordCount() const { return this->texCoordCount; };

  /** @returns the Count of Faces of this Model */
  inline unsigned int getFaceCount() const { return this->faceCount; };


  Material* addMaterial(Material* material);
  Material* addMaterial(const char* materialName);


  bool addVertex(float x, float y, float z);

  bool addVertexNormal(float x, float y, float z);

  bool addVertexTexture(float u, float v);

  bool addFace(int faceElemCount, VERTEX_FORMAT type, ...);

  bool addGroup(const char* groupString);

  bool setMaterial(Material* mtl);

  void finalize();


 private:
  bool buildTriangleList();

  bool addGLElement(ModelFaceElement* elem);

  bool importToVertexArray();

  bool deleteArrays();
  bool cleanup();

 private:
  bool                       finalized;       //!< Sets the Object to be finalized.

  tArray<GLfloat>            vertices;        //!< The Array that handles the Vertices.
  tArray<GLfloat>            normals;         //!< The Array that handles the Normals.
  tArray<GLfloat>            textures;        //!< The Array that handles the VertexTextureCoordinates.

  tArray<GLfloat>            indices;         //!< The Array that tells us what Vertex is connected to which other one.

  std::list<ModelMaterial*>  materialList;    //!< A list for all the Materials in this Model
};

#endif
